package CalcRules;

import java.io.File;
import java.io.FileNotFoundException;
import java.util.Arrays;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import java.util.Scanner;


public class Input {
    public int support,confidence;    
    public String dataFileLocation, decisionAttribute, decisionFrom, decisionTo;    
    public List<String> attributeNames = new ArrayList<String>();
    public List<String> stableAttributes = new ArrayList<String>();
    public List<String> flexibleAttributes = new ArrayList<String>();
    public ArrayList<ArrayList<String>> data = new ArrayList<ArrayList<String>>();

    public void setDataFileLocation(String dataFilePath) {
        this.dataFileLocation = dataFilePath;
        readAttributesData();
    }

    public void readAttributesData() {
        String[] row;
        for (String attribute:attributeNames) {
            data.add(new ArrayList<String>());
        }
        try {
            Scanner input = new Scanner(new File(dataFileLocation));

            while (input.hasNextLine()) {
                row = input.nextLine().split(",");

                for (int i=0; i<row.length; i++) {
                    data.get(i).add(row[i].trim());

                }
            }
        } catch (FileNotFoundException e) {
            printMessage("File Not Found");
            e.printStackTrace();
        }
    }

    public void setSupport(int support) {
        this.support = support;
    }

    public void setConfidence(int confidence) {
        this.confidence = confidence;
    }

    public void setAttributeNames(String[] attributeNames) {
        this.attributeNames = new ArrayList<String>(Arrays.asList(attributeNames));
        this.flexibleAttributes = new ArrayList<String>(Arrays.asList(attributeNames));
    }

    public void setDecisionAttribute(String decisionAttribute) {
        this.decisionAttribute = decisionAttribute;
        this.flexibleAttributes.remove(decisionAttribute);
    }

    public void setStableAttribute(int[] stableAttributesIndex) {
        for (int attributeIndex:stableAttributesIndex) {
            this.stableAttributes.add(this.attributeNames.get(attributeIndex));
        }

        for (String attribute:this.stableAttributes) {
            if(this.flexibleAttributes.contains(attribute)) {
                this.flexibleAttributes.remove(attribute);
            }
        }
    }

    public void setDecisionFromValue(String decisionFrom)
    {
        int decisionAttributeIndex = attributeNames.indexOf(decisionAttribute);

        if (data.get(decisionAttributeIndex).contains(decisionFrom)) {
            this.decisionFrom = decisionFrom;
        }
        else {
            printMessage("Invalid value.");
        }
    }

    public void setDecisionToValue(String decisionTo)
    {
        int decisionAttributeIndex = attributeNames.indexOf(decisionAttribute);

        if (data.get(decisionAttributeIndex).contains(decisionTo)) {
            this.decisionTo = decisionTo;
        }
        else {
            printMessage("Invalid value.");
        }
    }

    public ArrayList<ArrayList<String>> getData() { return data; }
    public int getSupport() { return support; }
    public int getConfidence() { return confidence; }
    public List<String> getAttributeNames() { return attributeNames; }
    public String getDecisionAttribute() { return decisionAttribute; }
    public List<String> getStableAttributes() { return stableAttributes; }
    public List<String> getFlexibleAttributes() { return flexibleAttributes; }
    public String getDecisionFromValue() { return decisionFrom; }
    public String getDecisionToValue() { return decisionTo; }

    public void printMessage(String content) {
        System.out.println(content);
    }

    public void printList(List<String> list) {
        Iterator iterate = list.iterator();

        while(iterate.hasNext()) {
            printMessage(iterate.next().toString());
        }
    }
}